package com.db4o.tools;

import java.io.*;

import com.db4o.*;
import com.db4o.ext.*;

/**
 * defragments a database file.
 * <br><br>This class is not part of db4o.jar. It is delivered as sourcecode
 * in the path ../com/db4o/tools/<br><br>
 * <b>Prerequites:</b><br>
 * - The database file may not be in use.<br>
 * - All stored classes need to be available.<br>
 * - If you use special Db4o translators, they need to be installed
 * with<br><code>Db4o.configure().objectClass("classname").<br>
 *  translate(new Translatorclass());
 * </code><br>before starting Defragment.<br><br>
 * <b>Performed tasks:</b><br>
 * - Free filespace is removed.<br>
 * - Deleted IDs are removed.<br>
 * - Unavailable classes are removed.<br>
 * - Unavailable class members are removed.<br>
 * - Class indices are restored.<br>
 * - Previous rename tasks are removed.<br>
 * <br>
 * <b>Backup:</b><br>
 * Defragment creates a backup file with the name [filename].bak. If
 * a file with this name is already present, Defragment will not run
 * for safety reasons.<br><br>
 * <b>Recommendations:</b><br>
 * - Keep the backup copy of your database file.<br>
 * - Scan the output log for "Class not available" messages.
 */
public class Defragment
{
	/**
	 * the main method is the only entry point
	 */
	public Defragment(){}

	/**
	 * the main method that runs Defragment.
	 * @param String[] a String array of length 1, with the name of the database
	 * file as element 0.
	 */
	public static void main(String[] args){
		if(args == null || args.length != 1){
			System.out.println("Usage: java com.db4o.tools.Defragment <database filename>");
		}else{
			new Defragment().run(args[0]);
		}
	}

	private void run(String filename){
		File file = new File(filename);
		if(file.exists()){
			ExtFile backupTest = new ExtFile(file.getAbsolutePath() + ".bak");
			if(backupTest.exists()){
				System.out.println("A backup file with the name ");
				System.out.println("'" + backupTest.getAbsolutePath() + "'");
				System.out.println("already exists.");
				System.out.println("Remove this file before calling 'Defragment'.");
			}else{
				file.renameTo(backupTest);
				try{
					configureDb4o();
					ObjectContainer readFrom = Db4o.openFile(backupTest.getAbsolutePath());
					ObjectContainer writeTo = Db4o.openFile(file.getAbsolutePath());
					migrate(readFrom, writeTo);
					readFrom.close();
					writeTo.close();
					System.out.println("Defragment operation completed successfully.");
				}catch(Exception e){
					System.out.println("Defragment operation failed.");
					e.printStackTrace();
					try{
						new File(filename).delete();
						backupTest.copy(filename);
					}catch(Exception ex){
						System.out.println("Restore failed.");
						System.out.println("Please use the backup file:");
						System.out.println("'" + backupTest.getAbsolutePath() + "'");
						return;
					}
					System.out.println("The original file was restored.");
					try{
						new File(backupTest.getAbsolutePath()).delete();
					}catch(Exception ex){}
				}
			}
		}else{
			System.out.println("File '" + file.getAbsolutePath() + "' does not exist.");
		}
	}

	private void configureDb4o(){
		Db4o.configure().activationDepth(0);
		Db4o.configure().callbacks(false);
		Db4o.configure().classActivationDepthConfigurable(false);
		Db4o.configure().setOut(System.out);
	}

	private void migrate(ObjectContainer origin, ObjectContainer destination){
		// all Internal interfaces are undocumented and
		// may be subject to change without notice
		StoredClass[] classes = origin.ext().storedClasses();
		for(int i=0; i < classes.length; i++){
			long[] ids = classes[i].getIDs();
			origin.ext().purge();
			destination.commit();
			destination.ext().purge();
			for(int j=0; j < ids.length; j++){
				Object obj = origin.ext().getByID(ids[j]);

				// prevent possible constructor side effects
				origin.activate(obj, 1);
				origin.deactivate(obj, 2);

				origin.activate(obj, 1);
				destination.set(obj);

				// Both Containers keep track of state individually,
				// so we need to make sure, both know, the object is deactivated
				origin.deactivate(obj,1);
				destination.deactivate(obj,1);
			}
		}
	}
	
	
	private class ExtFile extends File{
		
		public ExtFile(String path){
			super(path);
		}
	
		public ExtFile copy(String toPath) throws Exception{
			try{
				new ExtFile(toPath).mkdirs();
				new ExtFile(toPath).delete();
				final int bufferSize = 64000;
				
				RandomAccessFile rafIn = new RandomAccessFile(getAbsolutePath(),"r");
				RandomAccessFile rafOut = new RandomAccessFile(toPath,"rw");
				long len = rafIn.length();
				byte[] bytes = new byte[bufferSize];
				
				while(len > 0){
					len -= bufferSize;
					if(len < 0){
						bytes = new byte[(int)(len + bufferSize)];	
					}
					rafIn.read(bytes);
					rafOut.write(bytes);
				}
				rafIn.close();
				rafOut.close();
				return new ExtFile(toPath);
			}
			catch(Exception e){
				e.printStackTrace();
				throw e;
			}
		}
	}
	
}

